#include "stdafx.h"
#include <math.h>
#include "FMOD.h"

/******************************************************/
/*            sound					                          */

Memx<FM::soundPtr> FM::Sounds;

FM::Sound& FM::Sound::del() {
	// this really removes the fmod channel pointer
	// the sound is kept in memory for later use
	Fader.remove(_channel);
	_channel->stop();
	_channel = NULL;
	isReady = false;
	return T;
}

FM::Sound::~Sound() {
	del();
	if (_listPtr) Sounds.removeData(_listPtr);
}

FM::Sound::Sound() {
	_channel=NULL; 
	isReady = false;
	_deleteAtZero = false;
	_fadeTarget = _fadeCurrent = 1;
	_blockTarget = _blockCurrent = 1;
	_listPtr = NULL;
}

FM::Sound& FM::Sound::create(Str name, Bool loop, Flt volume, FMOD::ChannelGroup * channelgroup, bool streaming, bool blocking) {
	// remember these
	_blocking				= blocking		;
	_loading_loop		= loop				;	
	_loading_volume = volume			;
	_loading_group	= channelgroup;
	_streaming			= streaming		;
	_playLoops = loop;
	
	// add this sound to pool and get soundpool position
	if (streaming) {
		_ID = FM::SoundSystem.addStream(name);
		// blocking should always be false for streams
		_blocking = false;
	} else {
		if (_blocking) _ID = FM::SoundSystem.addSound(name);
		else _ID = FM::SoundSystem.addSoundNB(name);
	}

	_currentSound = FM::SoundSystem.getSoundPtr(_ID);
	_is3D					= false;
	_speed				= 1.0	 ;
	
	if (_currentSound != NULL) {
		if (!_blocking) {
			FM::SoundSystem.addLoadingSound(this);
			isReady = false;
		} else {
			_setChannel();
			isReady = true;
		}
	}

	soundPtr & p = Sounds.New();
	p(this);
	_listPtr = &p;
	return T;
}

FM::Sound& FM::Sound::create(Str name, C Vec &pos, Flt range, Bool loop, Flt volume, FMOD::ChannelGroup * channelgroup, bool streaming, bool blocking) {
	// remember these
	_blocking				= blocking		;
	_loading_loop		= loop				;
	_loading_volume = volume			;
	_loading_group	= channelgroup;
	_loading_range	= range				;
	_loading_pos		= pos					;
	_streaming			= streaming		;
	_playLoops = loop;

	// add this sound to pool and get soundpool position
	if (streaming) {
		_ID = FM::SoundSystem.addStream(name);
		// blocking should always be false for streams
		_blocking = false;
	} else {
		if (_blocking) _ID = FM::SoundSystem.addSound(name);
		else _ID = FM::SoundSystem.addSoundNB(name);
	}

	_currentSound = FM::SoundSystem.getSoundPtr(_ID);
	_is3D					= true;
	_speed				= 1.0	;
	
	if (_currentSound != NULL) {
		if (!_blocking) {
			FM::SoundSystem.addLoadingSound(this);
			isReady = false;
		} else {
			_setChannel();
			isReady = true;
		}
	}

	soundPtr & p = Sounds.New();
	p(this);
	_listPtr = &p;
	return T;
}

void FM::Sound::update() {
	if (!isReady) return;

	if (_fadeCurrent == 0 && _deleteAtZero) {
		del();
		return;
	}

	if (!_playLoops) {
		if (time() == length()) pause();
	}

	Bool change = false;

	// fade changes can take as long as you wish
	if (_fadeCurrent > _fadeTarget) {
		_fadeCurrent -= (1.0 / Abs(_fadeStep)) * Time.rd();
		_fadeCurrent < _fadeTarget ? _fadeCurrent = _fadeTarget : NULL ;
		change = true;
	} else if (_fadeCurrent < _fadeTarget) {
		_fadeCurrent += (1.0 / Abs(_fadeStep)) * Time.rd();
		_fadeCurrent > _fadeTarget ? _fadeCurrent = _fadeTarget : NULL ;
		change = true;
	}

	// blocking a sound should be very fast (200 Ms), but still slow enough to avoid popping
	if (_blockCurrent > _blockTarget) {
		_blockCurrent -= 5.0 * Time.rd();
		_blockCurrent < _blockTarget ? _blockCurrent = _blockTarget : NULL ;
		change = true;
	} else if (_blockCurrent < _blockTarget) {
		_blockCurrent += 5.0 * Time.rd();
		_blockCurrent > _blockTarget ? _blockCurrent = _blockTarget : NULL ;
		change = true;
	}

	if (change) _channel->setVolume(_fadeCurrent * _blockCurrent);
}

void FM::Sound::_setChannel() {
	FM::SoundSystem.addChannel(_currentSound, _channel);
	if (_loading_group == NULL) _loading_group = ChannelFX();
	_channel->setChannelGroup(_loading_group);

	_channel->setLoopCount(-1);

	_channel->setVolume(_loading_volume);
	_fadeTarget = _fadeCurrent = _loading_volume;

	if (_is3D) {
		_channel->setMode(FMOD_3D);
		// add 3D specific stuff
		_channel->set3DMinMaxDistance(_loading_range, 10000.0);
		FMOD_VECTOR position = FM::vecToFmod(_loading_pos);
		FMOD_VECTOR velocity;
		_lastPos = position;
		velocity.x = 0; velocity.y = 0; velocity.z = 0;
		_channel->set3DAttributes(&position, &velocity);

		// spread stereo sounds in 3D
		int channels;
		_currentSound->getFormat(NULL, NULL, &channels, NULL);
		if (channels > 1) {
			Flt spread = channels * 45;
			// limit this a bit
			if (spread > 270) spread = 270;
			_channel->set3DSpread(spread);
		}
	} else { // 2D sound
		_channel->setMode(FMOD_2D);
	}

	// get channel name for save file
	char channel_name[100];
	_loading_group->getName(channel_name, 100);
	_channel_group = _loading_group;

	// start playing if needed
	if (_playing) play();
};
	

FM::Sound& FM::Sound::stop() {
	if (isReady) {
		_channel->setPaused(true);
		// internet streams cannot reposition
		if (!_streaming) _channel->setPosition(0, FMOD_TIMEUNIT_MS);
	}
	_playing = false;
	return T;
}

FM::Sound& FM::Sound::pause() {
	if (isReady) _errorCheck(T.name(), _channel->setPaused(true));
	_playing = false;
	return T;
}

Bool FM::Sound::play() {
	if (isReady) {
		_errorCheck(T.name(), _channel->setPaused(false));
		_playing = true;
		return true;
 	} else return false;
}

Bool FM::Sound::playing() {
	// we don't use isPlaying here for good reasons!
  if (isReady) {
	  Bool paused;
	  _errorCheck(T.name(), _channel->getPaused(&paused));
	  return !paused;
  }
  return false;
}

Str FM::Sound::name() {
	return FM::SoundSystem.getSoundName(_ID);
}

Int FM::Sound::raws() {
	// should this be PCMBYTES or RAWBYTES?
	// RAWBYTES cannot be used to set a raw position, so 
	// I'll choose PCMBYTES for now
	UInt length;
	_currentSound->getLength(&length, FMOD_TIMEUNIT_PCMBYTES);
	return (signed)length;
}

Flt FM::Sound::length() {
	UInt length;
	_currentSound->getLength(&length, FMOD_TIMEUNIT_MS);
	return length / 1000.0;
}

FM::Sound& FM::Sound::raw(Int raw) {
	_channel->setPosition((unsigned)raw, FMOD_TIMEUNIT_PCMBYTES);
	return T;
}

Int FM::Sound::raw() {
	UInt raw;
	_channel->getPosition(&raw, FMOD_TIMEUNIT_PCMBYTES);
	return (signed)raw;
}

FM::Sound& FM::Sound::time(Flt t) {
	UInt pos = (UInt)floor(t * 1000);
	_errorCheck(S + "time function, " + T.name(), _channel->setPosition(pos, FMOD_TIMEUNIT_MS));
	return T;
}

Flt FM::Sound::time() {
	UInt pos;
	_channel->getPosition(&pos, FMOD_TIMEUNIT_MS);
	return pos / 1000.0;
}

FM::Sound& FM::Sound::frac(Flt f) {
	UInt length;
	_currentSound->getLength(&length, FMOD_TIMEUNIT_MS);
	length /= f;
	_channel->setPosition(length, FMOD_TIMEUNIT_MS);
	return T;
}

Bool FM::Sound::_ready() {
	FMOD_OPENSTATE result;
	FMOD_RESULT error = _currentSound->getOpenState(&result, NULL, NULL, NULL);
	//_errorCheck(T.name(), error);
	if ((result == FMOD_OPENSTATE_READY)||(result == FMOD_OPENSTATE_PLAYING)) return true;
	return false;
}


Flt FM::Sound::frac() {
	UInt length;
	_currentSound->getLength(&length, FMOD_TIMEUNIT_MS);
	UInt pos;
	_channel->getPosition(&pos, FMOD_TIMEUNIT_MS);
	return pos / length;
}

FM::Sound& FM::Sound::loop(Bool loop) {
	if (loop) _channel->setLoopCount(-1);
	else _channel->setLoopCount(0);
	return T;
}

Bool FM::Sound::loop() {
	Int count;
	_channel->getLoopCount(&count);
	if (count == 0) return false;
	return true;
}

FM::Sound& FM::Sound::volume(Flt volume) {
	// setting volume is just a very short fade
	_fadeTarget = volume;
	_fadeStep = (_fadeTarget - _fadeCurrent) / 0.2;
	return T;
}

Flt FM::Sound::volume() {
	//Flt volume;
	//_channel->getVolume(&volume);
	return _fadeTarget;
}

FM::Sound& FM::Sound::speed(Flt speed) {
	_speed = speed;
	Flt frequency;
	_errorCheck(S + T.name() + ": set speed : " , _currentSound->getDefaults(&frequency, NULL, NULL, NULL));
	frequency *= _speed;
	_channel->setFrequency(frequency);
	return T;
}

Flt FM::Sound::speed() {
	return _speed;
}

FM::Sound& FM::Sound::spread(Flt spread) {
	Clamp(spread, 0, 360);
	_channel->set3DSpread(spread);
	return T;
}

Flt FM::Sound::spread() {
	Flt s;
	_channel->get3DSpread(&s);
	return s;
}

FM::Sound& FM::Sound::range(Flt range) {
	if (_is3D) _channel->set3DMinMaxDistance(range, 10000.0);
	return T;
}

Flt FM::Sound::range() {
	if (_is3D) {
		Flt range;
		_channel->get3DMinMaxDistance(&range, NULL);
		return range;
	}
	return 0;
}

FM::Sound& FM::Sound::pos(C Vec &pos) {
	Bool playing;
	_channel->isPlaying(&playing);
	if(playing && _is3D) {
		FMOD_VECTOR position = FM::vecToFmod(pos);
		FMOD_VECTOR velocity;
		velocity.x = (position.x - _lastPos.x) * (1 / Time.ad());
		velocity.y = (position.y - _lastPos.y) * (1 / Time.ad());
		velocity.z = (position.z - _lastPos.z) * (1 / Time.ad());
		_errorCheck(S + T.name() + " : set3D : " , _channel->set3DAttributes(&position, &velocity));
		_lastPos = position;
	}
	return T;
}

Vec FM::Sound::pos() {
	Vec pos;
	pos.set(_lastPos.x, _lastPos.y, _lastPos.z);
	return pos;
}

FM::Sound& FM::Sound::fadeInFromSilence(Flt fade_duration) {
	_channel->setVolume(0);
	_fadeCurrent = 0;
	_fadeTarget = 1;
	if(fade_duration < 0.1) fade_duration = 0.1;
	_fadeStep = (_fadeTarget - _fadeCurrent) / fade_duration;
	return T;
}

FM::Sound& FM::Sound::fadeIn(Flt fade_duration) {
	_fadeTarget = 1;
	if(fade_duration < 0.1) fade_duration = 0.1;
	_fadeStep = (_fadeTarget - _fadeCurrent) / fade_duration;
	return T;
}

FM::Sound& FM::Sound::fadeOut(Flt fade_duration, Bool deleteChannel) {
	_fadeTarget = 0;
	if(fade_duration < 0.1) fade_duration = 0.1;
	_fadeStep = (_fadeTarget - _fadeCurrent) / fade_duration;
	return T;
}

FM::Sound& FM::Sound::block() {
	_blockTarget = 0.2;
	return T;
}

FM::Sound& FM::Sound::unblock() {
	_blockTarget = 1;
	return T;
}

FM::Sound& FM::Sound::addDSP(FM::dsp & s) {
	dspPtr = &s;
	_channel->addDSP(dspPtr->dspPtr, NULL);
	return T;
}

void FM::Sound::save(File &f) {
	if(_channel != NULL) {
		f.putStr(FM::SoundSystem.getSoundName(_ID));
		f << _is3D;
		f << pos();
		f << _playing;
		f << raw();
		f << loop();
		f << volume();
		f << speed();
		f << range();
		f << _channel_group;
		f << _blocking;
		f << _streaming;
	}
}

Bool FM::Sound::load(File &f) {
	_channel->stop();
	Str name;
	Vec pos;
	Flt raw, volume, speed, range;
	Bool loop;
	f.getStr(name);
	f >> _is3D;
	f >> pos;
	f >> _playing;
	f >> raw;
	f >> loop;
	f >> volume;
	f >> speed;
	f >> range;
	f >> _channel_group;
	f >> _blocking;
	f >> _streaming;

	T.isReady = false;

	if (_is3D) {
		T.create(name, pos, range, loop, volume, FM::Channel(_channel_group), _streaming, _blocking);
	} else {
		T.create(name, loop, volume, FM::Channel(_channel_group), _streaming, _blocking);
	}
	T.raw(raw);
	T.speed(speed);
	if (!_blocking && _playing) T.play();

	return true;
}

void FM::SoundPlay(Str name, Flt volume, FMOD::ChannelGroup * channelgroup) {
	UInt _ID = FM::SoundSystem.addSound(name);
	FMOD::Sound * _currentSound = FM::SoundSystem.getSoundPtr(_ID);
	FMOD::Channel * _channel;
	if (_currentSound != NULL) {
		FM::SoundSystem.addChannel(_currentSound, _channel);
		if (channelgroup == NULL) channelgroup = ChannelFX();
		_channel->setChannelGroup(channelgroup);
		_channel->setLoopCount(0);
		_channel->setVolume(volume);
	}
	_channel->setPaused(false);
}
		
void FM::SoundPlay(Str name, C Vec &pos, Flt range, Flt volume, FMOD::ChannelGroup * channelgroup) {
	UInt _ID = FM::SoundSystem.addSound(name);
	FMOD::Sound * _currentSound = FM::SoundSystem.getSoundPtr(_ID);
	FMOD::Channel * _channel;
	if (_currentSound != NULL) {
		FM::SoundSystem.addChannel(_currentSound, _channel);
		if (channelgroup == NULL) channelgroup = ChannelFX();
		_channel->setChannelGroup(channelgroup);
		_channel->setLoopCount(0);
		_channel->setVolume(volume);
		_channel->setMode(FMOD_3D);
		_channel->set3DMinMaxDistance(range, 10000.0);
		_channel->set3DAttributes(&FM::vecToFmod(pos), 0);
		
		// spread stereo sounds in 3D
		int channels;
		_currentSound->getFormat(NULL, NULL, &channels, NULL);
		if (channels > 1) {
			Flt spread = channels * 45;
			// limit this a bit
			if (spread > 270) spread = 270;
			_channel->set3DSpread(spread);
		}
	}

	// start
	_channel->setPaused(false);
}

Bool FM::Sound::starving() {
	Bool _starving;
	_currentSound->getOpenState(NULL, NULL, &_starving, NULL);
	return _starving;
}

UInt FM::Sound::bufferFilled() {
	UInt _percent;
	_currentSound->getOpenState(NULL, &_percent, NULL, NULL);
	return _percent;
}


	







