/******************************************************************************
 * Copyright (c) yvan vander sanden.												                  *
 * Header file to wrap FMOD sound library in Esenthel Engine                  *
/******************************************************************************

   Use 'SoundVolume' to set  volume for different sound groups
   Use 'SoundPlay  ' to play custom sounds
   Use 'Sound      ' to play custom sounds and control them later
   Use 'Listener   ' to set  3D listener parameters which receives 3D sounds

/******************************************************************************/

#pragma once
#include "dsp.h"
#include "Channel.h"

namespace FM {

	struct ListenerControl {
		// get
		C Vec& pos() {return _pos;} // get listener position
		C Vec& dir() {return _dir;} // get listener looking direction
		C Vec& up () {return _up ;} // get listener up      direction
		C Vec& vel() {return _vel;} 

		// set
		FM::ListenerControl& orn(C Vec &dir, C Vec &up =Vec(0,1,0)); // set listener orientation, 'dir'=looking direction (must be normalized), 'up'=up direction (must be normalized)
		FM::ListenerControl& pos(C Vec &pos, C Vec &vel=Vec(0,0,0)); // set listener position and velocity

		FM::ListenerControl& mute(); // mute 3d sound listener
		void draw();

	private:
		Vec _pos,_vel,_dir,_up;

	};

	struct Sound;
	struct soundPtr;

	struct Sound {
		// manage
		FM::Sound&	del   ();
		FM::Sound&	create(Str name,                          Bool loop=0, Flt volume=1, FMOD::ChannelGroup * channelgroup = NULL, Bool streaming = false, Bool blocking = false);                                                                 // create             sound, 'name'=sound file name,                                                                       , 'loop'=if sound is looped, 'volume'=sound volume (0..1)
		FM::Sound&	create(Str name, C Vec &pos, Flt range=1, Bool loop=0, Flt volume=1, FMOD::ChannelGroup * channelgroup = NULL, Bool streaming = false, Bool blocking = false);                                                                 // create          3D sound, 'name'=sound file name, 'pos'=sound position in World, 'range'=sound range multiplier (0..Inf), 'loop'=if sound is looped, 'volume'=sound volume (0..1)
		
		// don't use play for sounds which take a while to load, because they won't be ready instantly
		Bool				play  (Str name,                          Bool loop=0, Flt volume=1, FMOD::ChannelGroup * channelgroup = NULL) {return create(name,          loop,volume,channelgroup, false, true).play();} // create and play    sound, 'name'=sound file name,                                                                       , 'loop'=if sound is looped, 'volume'=sound volume (0..1)
		Bool				play  (Str name, C Vec &pos, Flt range=1, Bool loop=0, Flt volume=1, FMOD::ChannelGroup * channelgroup = NULL) {return create(name,pos,range,loop,volume,channelgroup, false, true).play();} // create and play 3D sound, 'name'=sound file name, 'pos'=sound position in World, 'range'=sound range multiplier (0..Inf), 'loop'=if sound is looped, 'volume'=sound volume (0..1)

		// stop / play
		Sound& stop (); // stop
		Sound& pause(); // pause
		Bool   play (); // play

		// check if sound is ready (needed for non blocking sounds)
		Bool isReady;
		Bool	 _ready(); // check if sound is done loading (only do this from within system update, use bool isReady from elsewhere)
		
		// get / set
																		Bool playing(); // if playing
																		Str  name   (); //     get sound name
																		Int  raws   (); //     get sound raw  length (number of bytes)
																		Flt  length (); //     get sound time length (in seconds)
		Sound& raw   (  Int  raw   );   Int  raw    (); // set/get sound raw      position  , 0..raws  ()
		Sound& time  (  Flt  t     );   Flt  time   (); // set/get sound time     position  , 0..length()
		Sound& frac  (  Flt  f     );   Flt  frac   (); // set/get sound fraction position  , 0..1
		Sound& loop  (  Bool loop  );   Bool loop   (); // set/get sound looping            , true/false
		Sound& volume(  Flt  volume);   Flt  volume (); // set/get sound volume             , 0..1
		Sound& speed (  Flt  speed );   Flt  speed  (); // set/get sound speed              , 0.., default=1, negative speed plays the sound backwards
		Sound& spread(	Flt	 spread);		Flt	 spread	(); // set/get 3D spread in speaker space to 0..360 degrees
		// range with logarithmic rolloff (the default) is really the minimum distance before a sound
		// starts attenuating. This means it is at maximum volume until that point. 
		// See FMOD manual about logarithmic distance model for more about this.
		Sound& range (  Flt  range );   Flt  range  (); // set/get sound 3D range multiplier, 0..Inf
		Sound& pos   (C Vec &pos   );   Vec  pos    (); // set/get sound 3D position
		// velocity is calculated from last and new position
		//Sound& vel   (C Vec &vel   );   Vec  vel    (); // set/get sound 3D velocity

		// fade
		Sound& fadeInFromSilence(Flt fade_duration); // this method sets the fading value to silence, and then fades in the sound smoothly
		Sound& fadeIn           (Flt fade_duration); // fades in  the sound smoothly
		Sound& fadeOut          (Flt fade_duration, Bool deleteChannel = false); // fades out the sound smoothly
		
		// block (temporary block output, without changing volume or fader values)
		Sound& block();
		Sound& unblock();
		
		// dsp
		Sound& addDSP(FM::dsp & s);

		// io
		void save(File &f); // save sound settings (name, raw, pos, volume, ...)
		Bool load(File &f); // load sound settings (name, raw, pos, volume, ...), doesn't play the sound automatically

	 ~Sound();
		Sound();

		// update sound for smoothing of volume changes
		void update();

		// this is used by the struct itself or by system, don't use directly
		void _setChannel();
		
		// functions for streams
		Bool starving(); // for internet streams with not enough new data
		UInt bufferFilled(); // internet streams: examens the state of the streambuffer

	private:
		FMOD::Channel *		_channel			;
		FMOD::Sound		*		_currentSound	;
		FMOD_VECTOR				_lastPos			;
		UInt							_ID						;
		Bool							_is3D					;
		Flt								_speed				;
		Str								_channel_group;
		Bool							_blocking			;
		Bool							_playLoops		;
		Bool							_streaming		;
		Bool							_playing      ; // hold playing state for _setChannel()
		soundPtr *				_listPtr			;

		// for fading & blocking
		Bool _deleteAtZero;
		Flt _fadeTarget;
		Flt _fadeCurrent;
		Flt _fadeStep;
		Flt _blockTarget;
		Flt _blockCurrent;

		// dsp
		FM::dsp * dspPtr;

		// these are only used at loading time
		bool _loading_loop;
		Flt _loading_volume;
		FMOD::ChannelGroup * _loading_group;
		Vec _loading_pos;
		Flt _loading_range;
		
	};

/******************************************************************************/
	void SoundPlay(Str name,                          Flt volume=1, FMOD::ChannelGroup * channelgroup = NULL); // play    sound, 'name'=sound file name                                                                        , 'volume'=sound volume (0..1)
	void SoundPlay(Str name, C Vec &pos, Flt range=1, Flt volume=1, FMOD::ChannelGroup * channelgroup = NULL); // play 3D sound, 'name'=sound file name, 'pos'=sound position in World, 'range'=sound range multiplier (0..Inf), 'volume'=sound volume (0..1)
/******************************************************************************/

	struct soundPtr {
		Sound * ptr;
		Sound * operator()(						) {								return T.ptr; }
		Sound * operator()(Sound* ptr	) { T.ptr = ptr;	return T.ptr; }
	};

	extern SoundVolumeControl		SoundVolume	;
	extern ListenerControl			Listener		;
	extern Memx<soundPtr>				Sounds			; // system uses this for smooth volume adjustment
}



