#include "stdafx.h"
#include <string.h>
#include <string>
#include "FMOD.h"
#include "Sound.h"
#include "filesystem.h"

FM::System FM::SoundSystem;
Int FM::DSP::sampleRate;

/*******************************************************/
/*            System                                   */

FM::System::System() {
	_activated	= false;
	_underwater = false;
	_inCave			= false;

	// standard 3D settings
	_dopplerscale = 1.0;
	_distancefactor = 1.0;
	_rolloffscale = 1.0;

	// create soundsystem
	_errorCheck(System_Create(&_system));

	// find available drivers
	_system->getNumDrivers(&_driverCount);
	char name[256];
	for (Int x = 0; x < _driverCount; x++) {
		Str &elm = _drivers.New();
		_system->getDriverInfo(x, name, 256, NULL);
		elm = name;
	}
}

void FM::System::create(UInt driverID) {


	// check if we have the right version of the library
	UInt version;
	_errorCheck(_system->getVersion(&version));
	if (version < FMOD_VERSION) Exit("FMOD: outdated fmod dll.");

	// get the number of available audio drivers
	_errorCheck(_system->getNumDrivers(&_numDrivers));
	if (_numDrivers == 0) {
		// continue without sound
		_errorCheck(_system->setOutput(FMOD_OUTPUTTYPE_NOSOUND));
	} else {
		// revert to default driver if the ID is not right
		if (driverID > _numDrivers) driverID = 0;

		// Get driver info
		_errorCheck(_system->getDriverCaps(driverID, &_caps, 0, &_speakermode));

		// Set speakermode
		_errorCheck(_system->setSpeakerMode(_speakermode));
		if (_caps & FMOD_CAPS_HARDWARE_EMULATED) {
			_errorCheck(_system->setDSPBufferSize(1024, 10));
		}

		// something wrong with sigmatel audio drivers?
		_errorCheck(_system->getDriverInfo(driverID, _driverName, 256, 0));
		if(strstr(_driverName, "SigmaTel")) {
			_errorCheck(_system->setSoftwareFormat(48000, FMOD_SOUND_FORMAT_PCMFLOAT, 0, 0, FMOD_DSP_RESAMPLER_LINEAR));
		}
	}
	// try initializing the sound system
	FMOD_RESULT r = _system->init(1000, FMOD_INIT_NORMAL, 0);

	// if something went wrong, revert to play stereo mode
	if (r == FMOD_ERR_OUTPUT_CREATEBUFFER) {
		_errorCheck(_system->setSpeakerMode(FMOD_SPEAKERMODE_STEREO));
		_errorCheck(_system->init(1000, FMOD_INIT_NORMAL, 0));
	}

	// overwrite file callbacks for EE's data structure
	r = _system->setFileSystem(eeOpen, eeClose, eeRead, eeSeek, 0, 0, -1);

	_system->set3DSettings(_dopplerscale, _distancefactor, _rolloffscale);

	// now create known channel groups
	FM::ChannelGlobal.create();
	FM::ChannelFX.create();
	FM::ChannelAmbient.create();
	FM::ChannelMusic.create();
	FM::ChannelVoice.create();
	FM::ChannelGui.create();

	// add known musicmanagers
	addMusicManager(&FM::Music);
	addMusicManager(&FM::Ambient);

	// ready for action
	_activated = true;

	// just to make sure reverb is off
	FMOD_REVERB_PROPERTIES noVerb = FMOD_PRESET_OFF;
	_system->setReverbProperties(&noVerb);

	// initialize geometry
	_system->setGeometrySettings(4000.0);

	// get samplerate for dsp
	_system->getSoftwareFormat(&DSP::sampleRate, NULL, NULL, NULL, NULL, NULL);
}

UInt FM::System::addSound(C Str &fileName) {
	// check if we got this sound already
	Int pos = _findInPool(fileName);
	if (pos == -1) {
		// add this sound
		soundStruct tmp;
		tmp.name = fileName;
		Str fullName = S + fileName;
		_errorCheck(_system->createSound((Str8)fullName, (FMOD_MODE)(FMOD_SOFTWARE)|(FMOD_LOOP_NORMAL), 0, &tmp.sound));
		_soundPool.add(tmp);
		return _soundPool.find(tmp);
	} else {
		return pos;
	}
}

UInt FM::System::addSoundNB(C Str &fileName) {
	// check if we got this sound already
	Int pos = _findInPool(fileName);
	if (pos == -1) {
		// add this sound
		soundStruct tmp;
		tmp.name = fileName;
		Str fullName = S + fileName;
		_errorCheck(_system->createSound((Str8)fullName, (FMOD_MODE)(FMOD_SOFTWARE)|(FMOD_LOOP_NORMAL)|(FMOD_NONBLOCKING), 0, &tmp.sound));
		_soundPool.add(tmp);
		return _soundPool.find(tmp);
	} else {
		return pos;
	}
}

UInt FM::System::addStream(C Str& name) {
	// check if we got this sound already
	Int pos = _findInPool(name);
	if (pos == -1) {
		// add this sound
		soundStruct tmp;
		tmp.name = name;
		FMOD_RESULT result = _system->createStream((Str8)name, (FMOD_MODE)(FMOD_SOFTWARE)|(FMOD_LOOP_NORMAL)|(FMOD_NONBLOCKING), 0, &tmp.sound);
		if (result == FMOD_OK) {
			_soundPool.add(tmp);
			return _soundPool.find(tmp);
		} else {
			return -1;
		}
	} 
	return pos;
}

Int FM::System::_findInPool(C Str &fileName) {
	for (Int i = 0; i < _soundPool.elms(); i++) {
		if (Equal(_soundPool[i].name, fileName)) return i;
	}
	return -1;
}

Str FM::System::getSoundName(UInt ID) {
	if (ID > _soundPool.elms()) return "";
	return _soundPool[ID].name;
}

FMOD::Sound * FM::System::getSoundPtr(UInt ID) {
	if (ID > _soundPool.elms()) return NULL;
	return _soundPool[ID].sound;
}

void FM::System::addChannel(FMOD::Sound * sound, FMOD::Channel*& channel) {
	_errorCheck(_system->playSound(FMOD_CHANNEL_FREE, sound, true, &channel));
}

void FM::System::addChannelGroup(C Str &name, FMOD::ChannelGroup *& channel) {
	// Why can't i get a const char * from a EE::Str ???
	std::string n = "";
	for (int i = 0; i < name.length(); i++) {
		n += name[i];
	}
	_errorCheck(_system->createChannelGroup(n.c_str(), &channel));
}

void FM::System::addLoadingSound(FM::Sound * sound) {
	_loadingSounds.push_back(sound);
}

void FM::System::_createDSP(FMOD_DSP_TYPE type, FMOD::DSP ** dsp) {
	_system->createDSPByType(type, dsp);
}

void FM::System::_createDSP(FMOD_DSP_DESCRIPTION & description, FMOD::DSP ** dsp) {
	_errorCheck(_system->createDSP(&description, dsp));
}

void FM::System::_createReverb(FMOD::Reverb ** reverb) {
	_errorCheck(_system->createReverb(reverb));
}

void FM::System::addMusicManager(FM::MusicManager * manager) {
	_musicManagers.add(manager);
}

void FM::System::underwater(Bool status, Flt depth) {
	// lowpass filter on global channel
	FM::ChannelGlobal._lowpassFilter(status, depth);
	
	// reverb on system object
	if ((status)&&(!_underwater)) {
		FMOD_REVERB_PROPERTIES prop1 = FMOD_PRESET_UNDERWATER;
		_errorCheck(_system->setReverbAmbientProperties(&prop1));
		_underwater = true;
	} 
	if ((!status)&&(_underwater)) {
		if (_inCave) {
			FMOD_REVERB_PROPERTIES prop1 = FMOD_PRESET_CAVE;
			_errorCheck(_system->setReverbAmbientProperties(&prop1));
		} else {
			FMOD_REVERB_PROPERTIES prop1 = FMOD_PRESET_OFF;
			_errorCheck(_system->setReverbAmbientProperties(&prop1));
		}
		_underwater = false;
	}
}

void FM::System::insideCave(Bool status) {
	if (status == _inCave) return;

	if (status && !_inCave) {
		FMOD_REVERB_PROPERTIES prop1 = FMOD_PRESET_CAVE;
		_errorCheck(_system->setReverbAmbientProperties(&prop1));
		_inCave = true;
	} else {
		if (_underwater) {
			FMOD_REVERB_PROPERTIES prop1 = FMOD_PRESET_UNDERWATER;
			_errorCheck(_system->setReverbAmbientProperties(&prop1));
		} else {
			FMOD_REVERB_PROPERTIES prop1 = FMOD_PRESET_OFF;
			_errorCheck(_system->setReverbAmbientProperties(&prop1));
		}
		_inCave = false;
	}
}


void FM::System::update() {
	FM::Fader.update();

	// see if loading sounds are ready
	for (std::vector<FM::Sound *>::iterator pos = _loadingSounds.begin(); pos != _loadingSounds.end();) {
		FM::Sound * s = *pos;
		if (s->_ready()) {
			s->_setChannel();
			s->isReady = true;
			_loadingSounds.erase(pos);
		} else {
			pos++;
		}
	}

	// update sound volumes
	FREPA(Sounds) {
		Sounds[i]()->update();
	}

	FREPA(_musicManagers) {
		if (_musicManagers[i]->playing) {
			if ((_musicManagers[i]->length() - _musicManagers[i]->time()) 
				< _musicManagers[i]->fade_out) {
					_musicManagers[i]->next();
			}
		}
	}
	_system->set3DListenerAttributes( 0, 
		&FM::vecToFmod(FM::Listener.pos()),
		&FM::vecToFmod(FM::Listener.vel()),
		&FM::vecToFmod(FM::Listener.dir()),
		&FM::vecToFmod(FM::Listener.up())
	);
	_errorCheck(_system->update());
}

FM::System& FM::System::dopplerscale(Flt scale) {
	_dopplerscale = scale;
	if (_activated) _system->set3DSettings(_dopplerscale, _distancefactor, _rolloffscale);
	return T;
}

FM::System& FM::System::distancefactor(Flt factor) {
	_distancefactor = factor;
	if (_activated) _system->set3DSettings(_dopplerscale, _distancefactor, _rolloffscale);
	return T;
}

FM::System& FM::System::rolloffscale(Flt scale) {
	_rolloffscale = scale;
	if (_activated) _errorCheck(_system->set3DSettings(_dopplerscale, _distancefactor, _rolloffscale));
	return T;
}

Int FM::System::driverElms() {
	return _driverCount;
}

Str FM::System::driverName(Int ID) {
	if (ID < _driverCount) {
		return _drivers[ID];
	} else {
		return "";
	}
}

Int FM::System::driverID(Str name) {
	for (Int x = 0; x < _driverCount; x++) {
		if (Equal(name, _drivers[x])) {
			return x;
		}
	}
	// if not found
	return -1;
}

void FM::System::driver(Str name) {
	Int ID = driverID(name);
	if (ID > -1) _errorCheck(_system->setDriver(ID));
}

void FM::System::driver(Int ID) {
	if (ID > -1) _errorCheck(_system->setDriver(ID));
}

FM::System::~System() {
	_system->release();
}

FM::geometryObject& FM::System::createGeometry(PhysBody body, Flt directOcclusion, Flt reverbOcclusion, Bool doubleSided) {
	// construct a temporary container with simplified meshes
	// for all parts of this object
	Memc<MeshBase> base;
	FREPAD(k, body.parts) {
		//if (body.parts[k].type() == PHYS_MESH) {
			MeshBase& newBase = base.New();
			newBase.create(body.parts[k]);
			newBase.triToQuad();
			newBase.removeUnusedVtxs(false);
			newBase.removeDegenerateFaces();
		//}
	}

	// see how many vertexes and polygons we have
	Int vertexes = 0;
	Int polygons = 0;
	FREPAD(x, base) {
		vertexes += (base[x].tris() * 3);
		vertexes += (base[x].quads() * 4);
		polygons += base[x].faces();
	}

	// create a new geometry object
	FM::geometryObject& obj = _geometryObjects.New();
	_errorCheck(_system->createGeometry(polygons, vertexes, &obj._object));

	// add polygons to geometry object
	FREPAD(x, base) {
		// iterate through all triangles
		for (Int i = 0; i < base[x].tris(); i++) {
			VecI trianglePoints = base[x].tri.ind(i);
			// construct a vector for these points
			FMOD_VECTOR point[3];
			point[0] = FM::vecToFmod(base[x].vtx.pos(trianglePoints.x));
			point[1] = FM::vecToFmod(base[x].vtx.pos(trianglePoints.y));
			point[2] = FM::vecToFmod(base[x].vtx.pos(trianglePoints.z));

			Int index; // we won't save the index since it is not needed
			
			// add this polygon to the geometry object
			_errorCheck(obj._object->addPolygon(
				directOcclusion, 
				reverbOcclusion, 
				doubleSided, 
				3, 
				point, 
				&index)
			);
		}

		// now do the same for quads
		for (Int i = 0; i < base[x].quads(); i++) {
			VecI4 quadPoints = base[x].quad.ind(i);
			// construct a vector for these points
			FMOD_VECTOR point[4];
			point[0] = FM::vecToFmod(base[x].vtx.pos(quadPoints.x));
			point[1] = FM::vecToFmod(base[x].vtx.pos(quadPoints.y));
			point[2] = FM::vecToFmod(base[x].vtx.pos(quadPoints.z));
			point[3] = FM::vecToFmod(base[x].vtx.pos(quadPoints.w));

			Int index; // we won't save the index since it is not needed
			
			// add this polygon to the geometry object
			_errorCheck(obj._object->addPolygon(
				directOcclusion, 
				reverbOcclusion, 
				doubleSided, 
				4, 
				point, 
				&index)
			);
		}
	}

	return obj;
}

Flt FM::System::directOcclusion(Vec listener, Vec sound) {
	Flt result;
	_system->getGeometryOcclusion(
		&FM::vecToFmod(listener), &FM::vecToFmod(sound),
		&result, 0
	);
	return result;
}

Flt FM::System::reverbOcclusion(Vec listener, Vec sound) {
	Flt result;
	_system->getGeometryOcclusion(
		&FM::vecToFmod(listener), &FM::vecToFmod(sound),
		0, &result
	);
	return result;
}

void FM::_errorCheck(C FMOD_RESULT &result) {
	//if (!WORK_COPY) return;
	if (result != FMOD_OK) {
		Str txt = S + FMOD_ErrorString(result);
		//Lg(txt);
		Exit(txt);
	}
}

void FM::_errorCheck(C Str message, C FMOD_RESULT &result) {
	//if (!WORK_COPY) return;
	if (result != FMOD_OK) {
		Str txt = S + message + ": " + FMOD_ErrorString(result);
		//Lg(txt);
		Exit(txt);
	}
}

void FM::System::lockDSP() {
	_system->lockDSP();
}

void FM::System::unlockDSP() {
	_system->unlockDSP();
}
