/******************************************************************************
 * Copyright (c) yvan vander sanden.												                  *
 * Header file to wrap FMOD sound library in Esenthel Engine                  *
/******************************************************************************/

#pragma once

#include "Sound.h"
#include "SoundStream.h"
#include "Internals.h"
#include <vector>

namespace FM {
	// this is used by dsp functions
	namespace DSP {
		extern Int sampleRate;
	}
	struct MusicManager;
	struct geometryObject;

	struct System {
		void create	(UInt		driverID = 0); // driverID will revert to 0 if not valid, call at game init
		void update	(										); // call this during game update
		
		// use underwater reverb and lowpassfilter on global channel
		void underwater(Bool status, Flt depth = 0);

		// use cave reverb on system object
		void insideCave(Bool status);

		// add a new musicmanager, so the update system can find it
		void addMusicManager(FM::MusicManager * manager);

		// get/set 3D settings
		System& dopplerscale		(Flt scale	); Flt dopplerscale		() { return _dopplerscale		; }
		System& distancefactor	(Flt factor	); Flt distancefactor	() { return _distancefactor	; }
		System& rolloffscale		(Flt scale	); Flt rolloffscale		() { return _rolloffscale		; }

		// for choosing an audiodriver
		Int driverElms();
		Meml<Str> & getDriverList() { return _drivers; }
		Str driverName(Int ID);
		Int driverID(Str name);
		void driver(Str name); void driver(Int ID);

		// add a geometry object
		geometryObject& createGeometry(PhysBody object, Flt directOcclusion = 1.0f, Flt reverbOcclusion = 1.0f, Bool doubleSided = false);
		
		// use these to check on occlusion between two points
		Flt directOcclusion(Vec listener, Vec sound);
		Flt reverbOcclusion(Vec listener, Vec sound);

		// required if you change the values of any dsp object
		void lockDSP();
		void unlockDSP();

		/**********************************/
		/* following are for internal use */
		/**********************************/

		// for sounds
		UInt					addSound		(C Str	&fileName	); // make this file available to the system, you should not need to call this directly. Returns internal ID of the sound
		UInt					addSoundNB	(C Str  &fileName ); // same but nonblocking, test if sound is ready before trying to access it.
		UInt					addStream		(C Str& name			); // stream is always nonblocking
		Str						getSoundName(UInt ID					);
		FMOD::Sound * getSoundPtr	(UInt ID					);

		

		// for internal use: this is for real channels, i.e. groups of playing sounds
		void addChannelGroup(C Str& name, FMOD::ChannelGroup *& channel);

		// for internal use: a channel is a playing sound - every playing sound has its own channel
		void addChannel(FMOD::Sound * sound,	FMOD::Channel*& channel);
		
		// add nonblocking sound to queue while loading
		void addLoadingSound(FM::Sound * sound);

		Bool active() { return _activated; }
		void _createDSP(FMOD_DSP_TYPE type, FMOD::DSP ** dsp);
		void _createDSP(FMOD_DSP_DESCRIPTION &description, FMOD::DSP ** dsp);
		void _createReverb(FMOD::Reverb ** reverb);

		// destructor releases all resources
		System	();
		~System	();

	private:
		// FMOD internal needs
		FMOD::System *		_system						;
		Int								_numDrivers				;
		FMOD_SPEAKERMODE	_speakermode			;
		FMOD_CAPS					_caps							;
		char							_driverName[256]	;

		// if soundsystem is initialized
		Bool _activated;

		// container for all loaded sounds
		Memc<FM::soundStruct>		_soundPool		;
		Memc<FM::MusicManager*> _musicManagers;

		Int		_findInPool(C Str					&fileName	);

		// for underwater effect
		Bool	_underwater	;
		Flt		_depth			;

		// for cave effect
		Bool _inCave;

		// list with audio drivers
		Meml<Str> _drivers;
		Int _driverCount;

		// 3D settings
		Flt _dopplerscale;
		Flt _distancefactor;
		Flt _rolloffscale;

		// container for geometry objects
		Meml<geometryObject> _geometryObjects;
		// container for loading nonblocking sounds
		std::vector<FM::Sound*> _loadingSounds;
	};
	
	extern System SoundSystem;

	void	_errorCheck(C FMOD_RESULT	&result		);
	void _errorCheck(C Str message, C FMOD_RESULT	&result		);
}

