#include "stdafx.h"
#include "filters.h"

FM::DSP::highPass::highPass() {
	last = 0;
}

FM::DSP::highPass& FM::DSP::highPass::setFrequency(Flt f) {
	if (f < 0) f = 0;
	freq = f;
	coef = 1 - f * (2 * 3.14159) / sampleRate;
	if (coef < 0) coef = 0;
	else if (coef > 1) coef = 1;

	return T;
}

FM::DSP::samplePtr& FM::DSP::highPass::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());

	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	if (coef < 1) {
		for (UInt i = 0; i < length; i++) {
			Flt f = *inPtr++ + coef * last;
			*outPtr++ = f - last;
			last = f;
		}
	} else {
		last = 0;
		return in;
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::lowPass::lowPass() {
	last = 0;
}

FM::DSP::lowPass& FM::DSP::lowPass::setFrequency(Flt f) {
	if (f < 0) f = 0;
	freq = f;
	coef = 1 - f * (2 * 3.14159) / sampleRate;
	if (coef < 0) coef = 0;
	else if (coef > 1) coef = 1;

	return T;
}

FM::DSP::samplePtr& FM::DSP::lowPass::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());

	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	Flt feedback = 1 - coef;
	for (Int i = 0; i < length; i++) {
		last = *outPtr++ = coef * *inPtr++ + feedback * last;
	}

	return buffer;
}

/*******************************************************************************************/


FM::DSP::bandPass& FM::DSP::bandPass::set(Flt freq, Flt q) {
	T.freq = freq;
	T.q = q;
	calcCoef();
	return T;
}

FM::DSP::bandPass& FM::DSP::bandPass::setFrequency(Flt freq) {
	T.freq = freq;
	calcCoef();
	return T;
}

FM::DSP::bandPass& FM::DSP::bandPass::setQ(Flt q) {
	T.q = q;
	calcCoef();
	return T;
}

void FM::DSP::bandPass::calcCoef() {
	Flt r, oneminusr, omega;
	if (freq < 0.001) freq = 10;
	if (q < 0) q = 0;
	omega = freq * (2.0f * 3.14159f) / sampleRate;
	if (q < 0.001) oneminusr = 1.0f;
	else oneminusr= omega/q;
	if (oneminusr >  1.0f) oneminusr = 1.0f;
	r = 1.0f - oneminusr;
	coef1 = 2.0f * qCos(omega) * r;
	coef2 = -r * r;

	gain = 2 * oneminusr * (oneminusr + r * omega);
}

float FM::DSP::bandPass::qCos(Flt omega) {
	if (omega >= -(0.5 * 3.14159f) && omega <= 0.5f * 3.14159f) {
		Flt result = omega * omega;
		return (((result * result * result * (-1.0f/720.0f) + result * result * (1.0f / 24.0f)) - result * 0.5) + 1);
	} else return 0;
}

FM::DSP::bandPass::bandPass() {
	freq = 0;
	q = 0;
	last = previous = 0;
	calcCoef();
}

FM::DSP::samplePtr& FM::DSP::bandPass::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());

	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	for (Int i = 0; i < length; i++) {
		Flt output = *inPtr++ + coef1 * last + coef2 * previous;
		*outPtr++ = gain * output;
		previous = last;
		last = output;
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::biQuad::biQuad() {
	last = previous = 0;
	fb1 = fb2 = ff1 = ff2 = ff3 = 0;
}

FM::DSP::biQuad& FM::DSP::biQuad::setRaw(Flt fb1, Flt fb2, Flt ff1, Flt ff2, Flt ff3) {
	Flt discriminant = fb1 * fb1 + 4 * fb2;
	Bool zero = true;
	if (discriminant < 0) {
		if (fb2 >= -1.0f) zero = false;
	} else {
		if (fb1 <= 2.0f && fb1 >= -2.0f && 1.0f - fb1 - fb2 >= 0 && 1.0f + fb1 - fb2 >= 0) zero = false;
	}
	if (zero) fb1 = fb2 = ff1 = ff2 = ff3 = 0;
	T.fb1 = fb1;
	T.fb2 = fb2;
	T.ff1 = ff1;
	T.ff2 = ff2;
	T.ff3 = ff3;

	return T;
}

FM::DSP::samplePtr& FM::DSP::biQuad::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength()); 

	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	Flt output;
	for (Int i = 0; i < length; i++) {
		output = *inPtr++ + fb1 * last + fb2 * previous;
		*outPtr++ = ff1 * output + ff2 * last + ff3 * previous;
		previous = last;
		last = output;
	}

	return buffer;
}

FM::DSP::biQuad& FM::DSP::biQuad::set(BQ_TYPE type, Flt frequency, Flt Q, Flt peakGain) {
	T.type = type;
	T.freq = frequency;
	T.q = Q;
	T.peakGain = peakGain;

	calc();
	return T;
}

void FM::DSP::biQuad::calc() {
	Flt norm;
	Flt v = Pow(10, Abs(peakGain) / 20);
	Flt k = Tan(PI * freq / (Flt)sampleRate);

	switch (type) {
		case BQ_LOWPASS: {
			norm = 1 / (1 + k / q + k * k);
			ff1 = k * k * norm;
			ff2 = 2 * ff1;
			ff3 = ff1;
			fb1 = 2 * (k * k - 1) * norm;
			fb2 = (1 - k / q + k * k) * norm;
			break;
		}
		case BQ_HIGHPASS: {
			norm = 1 / (1 + k / q + k * k);
			ff1 = 1 * norm;
			ff2 = -2 * ff1;
			ff3 = ff1;
			fb1 = 2 * (k * k - 1) * norm;
			fb2 = (1 - k / q + k * k) * norm;
			break;
		}
		case BQ_BANDPASS: {
			norm = 1 / (1 + k / q + k * k);
			ff1 = k / q * norm;
			ff2 = 0;
			ff3 = -ff1;
			fb1 = 2 * (k * k - 1) * norm;
			fb2 = (1 - k / q + k * k) * norm;
			break;
		}
		case BQ_NOTCH: {
			norm = 1 / (1 + k / q + k * k);
			ff1 = (1 + k * k) * norm;
			ff2 = 2 * (k * k - 1) * norm;
			ff3 = ff1;
			fb1 = ff2;
			fb2 = (1 - k / q + k * k) * norm;
			break;
		}
		case BQ_PEAK: {
			if (peakGain >= 0) {
				norm = 1 / (1 + 1/q * k + k * k);
				ff1 = (1 + v/q * k + k * k) * norm;
				ff2 = 2 * (k * k - 1) * norm;
				ff3 = (1 - v/q * k + k * k) * norm;
				fb1 = ff2;
				fb2 = (1 - 1/q * k + k * k) * norm;
			} else {
				norm = 1 / (1 + v/q * k + k * k);
				ff1 = (1 + 1/q * k + k * k) * norm;
				ff2 = 2 * (k * k - 1) * norm;
				ff3 = (1 - 1/q * k + k * k) * norm;
				fb1 = ff2;
				fb2 = (1 - v/q * k + k * k) * norm;
			}
			break;
		}
		case BQ_LOWSHELF: {
			if (peakGain >= 0) {
				norm = 1 / (1 + SQRT2 * k + k * k);
				ff1 = (1 + Sqrt(2*v) * k + v * k * k) * norm;
				ff2 = 2 * (v * k * k - 1) * norm;
				ff3 = (1 - Sqrt(2*v) * k + v * k * k) * norm;
				fb1 = 2 * (k * k - 1) * norm;
				fb2 = (1 - SQRT2 * k + k * k) * norm;
			} else {	
				norm = 1 / (1 + Sqrt(2*v) * k + v * k * k);
				ff1 = (1 + SQRT2 * k + k * k) * norm;
				ff2 = 2 * (k * k - 1) * norm;
				ff3 = (1 - SQRT2 * k + k * k) * norm;
				fb1 = 2 * (v * k * k - 1) * norm;
				fb2 = (1 - Sqrt(2*v) * k + v * k * k) * norm;
			}
			break;
		}
		case BQ_HIGHSHELF: {
			if (peakGain >= 0) {
				norm = 1 / (1 + SQRT2 * k + k * k);
				ff1 = (v + Sqrt(2*v) * k + k * k) * norm;
				ff2 = 2 * (k * k - v) * norm;
				ff3 = (v - Sqrt(2*v) * k + k * k) * norm;
				fb1 = 2 * (k * k - 1) * norm;
				fb2 = (1 - SQRT2 * k + k * k) * norm;
			}
			else {	
				norm = 1 / (v + Sqrt(2*v) * k + k * k);
				ff1 = (1 + SQRT2 * k + k * k) * norm;
				ff2 = 2 * (k * k - 1) * norm;
				ff3 = (1 - SQRT2 * k + k * k) * norm;
				fb1 = 2 * (k * k - v) * norm;
				fb2 = (v - Sqrt(2*v) * k + k * k) * norm;
			}
			break;
	
		}
	}
	fb1 = -fb1;
	fb2 = -fb2;

	Flt discriminant = fb1 * fb1 + 4 * fb2;
	Bool zero = true;
	if (discriminant < 0) {
		if (fb2 >= -1.0f) zero = false;
	} else {
		if (fb1 <= 2.0f && fb1 >= -2.0f && 1.0f - fb1 - fb2 >= 0 && 1.0f + fb1 - fb2 >= 0) zero = false;
	}
	if (zero) fb1 = fb2 = ff1 = ff2 = ff3 = 0;
}

/*******************************************************************************************/

FM::DSP::sampleHold::sampleHold() {
	lastIn = lastOut = 0;
}

FM::DSP::samplePtr& FM::DSP::sampleHold::operator()(samplePtr& in, samplePtr& signal) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength()); 

	Flt * inPtr = in.getBuffer();
	Flt * sigPtr = signal.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	for (Int i = 0; i < length; i++, *inPtr++) {
		if (*sigPtr < lastIn) lastOut = *inPtr;
		*outPtr++ = lastOut;
		lastIn = *sigPtr++;
	}

	return buffer;
}



	



