#include "stdafx.h"
#include <math.h>
#include "math.h"

FM::DSP::samplePtr& FM::DSP::clip::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	while (length--) {
		Flt f = *inPtr++;
		if (f < low) f = low;
		if (f > high) f = high;
		*outPtr++ = f;
	}

	return buffer;
}

/*******************************************************************************************/

#define DUMTAB1SIZE 256
#define DUMTAB2SIZE 1024

float rsqrt_exptab[DUMTAB1SIZE], rsqrt_mantissatab[DUMTAB2SIZE];
Bool rSqrtInitOK = false;



void setupSqrt() {
	if (rSqrtInitOK) return;

	for (Int i = 0; i < DUMTAB1SIZE; i++) {
		Flt f;
		Int l = (i ? (i == DUMTAB1SIZE-1 ? DUMTAB1SIZE-2 : i) : 1)<< 23;
		*(Int *)(&f) = l;
		rsqrt_exptab[i] = 1./Sqrt(f);
	}

	for (Int i = 0; i < DUMTAB2SIZE; i++) {
		Flt f = 1 + (1./DUMTAB2SIZE) * i;
		rsqrt_mantissatab[i] = 1./Sqrt(f);
	}

	rSqrtInitOK = true;
}

FM::DSP::rSqrt::rSqrt() {
	setupSqrt();
}

FM::DSP::samplePtr& FM::DSP::rSqrt::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	while (length--) {
		Flt f = *inPtr;
		Int l = *(Int *)(inPtr++);
		if (f < 0) *outPtr++ = 0;
		else {
			Flt g = rsqrt_exptab[(l >> 23) & 0xff] * rsqrt_mantissatab[(l >> 13) & 0x3ff];
			*outPtr++ = 1.5 * g - 0.5 * g * g * g * f;
		}
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::sqrt::sqrt() {
	setupSqrt();
}

FM::DSP::samplePtr& FM::DSP::sqrt::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	while (length--) {
		Flt f = *inPtr;
		Int l = *(Int *)(inPtr++);
		if (f < 0) *outPtr++ = 0;
		else {
			Flt g = rsqrt_exptab[(l >> 23) & 0xff] * rsqrt_mantissatab[(l >> 13) & 0x3ff];
			*outPtr++ = f * (1.5 * g - 0.5 * g * g * g * f);
		}
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::wrap::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	while (length--) {
		Flt f = *inPtr++;
		Int k = f;
		if (f > 0) *outPtr++ = f - k;
		else *outPtr++ = f - (k-1);
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::midiToFreq::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	for (; length--; inPtr++, outPtr++) {
		Flt f = *inPtr;
		if (f < -1500) *outPtr = 0;
		else {
			if (f > 1499) f = 1499;
			*outPtr = 8.17579891564 * ::exp(.0577622650 * f);
		}
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::freqToMidi::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	for (; length--; *inPtr++, *outPtr++) {
		Flt f = *inPtr;
		*outPtr = (f > 0 ? 17.3123405046 * ::log(.12231220585 * f) : -1500);
	}

	return buffer;
}


/*******************************************************************************************/

#define LOGTEN 2.302585092994

FM::DSP::samplePtr& FM::DSP::dbToRms::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	for (; length--; *inPtr++, *outPtr++) {
		Flt f = *inPtr;
		if (f <= 0) *outPtr = 0;
		else {
			if (f > 485) f = 485;
			*outPtr = ::exp((LOGTEN * 0.05) * (f-100.));
		}
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::rmsToDb::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	for (; length--; *inPtr++, *outPtr++) {
		Flt f = *inPtr;
		if (f <= 0) *outPtr = 0;
		else {
			Flt g = 100 + 20./LOGTEN * ::log(f);
			*outPtr = (g < 0 ? 0 : g);
		}
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::dbToPow::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	for (; length--; *inPtr++, *outPtr++) {
		Flt f = *inPtr;
		if (f <= 0) *outPtr = 0;
		else {
			if (f > 870) f = 870;
			*outPtr = ::exp((LOGTEN * 0.1) * (f-100.));
		}
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::powToDb::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	for (; length--; *inPtr++, *outPtr++) {
		Flt f = *inPtr;
		if (f <= 0) *outPtr = 0;
		else {
			Flt g = 100 + 10./LOGTEN * ::log(f);
			*outPtr = (g < 0 ? 0 : g);
		}
	}

	return buffer;
}

/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::pow::operator()(samplePtr& in1, samplePtr& in2) {
	if (in1.getLength() != buffer.getLength()) buffer.resize(in1.getLength());
	Flt * in1Ptr = in1.getBuffer();
	Flt * in2Ptr = in2.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in1.getLength();

	while (length--) {
		Flt f = *in1Ptr++;
		if (f > 0) *outPtr = ::pow(f, *in2Ptr);
		else *outPtr = 0;
		outPtr++;
		in2Ptr++;
	}

	return buffer;
}


/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::exp::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	while (length--) *outPtr++ = ::exp(*inPtr++);

	return buffer;
}

/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::log::operator()(samplePtr& in1, samplePtr& in2) {
	if (in1.getLength() != buffer.getLength()) buffer.resize(in1.getLength());
	Flt * in1Ptr = in1.getBuffer();
	Flt * in2Ptr = in2.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in1.getLength();

	while (length--) {
		Flt f = *in1Ptr++, g = *in2Ptr++;
		if (f <= 0) *outPtr = ::log(f);
		else if (g <= 0) *outPtr = ::log(f); 
		else *outPtr = ::log(f)/::log(g);
		outPtr++;
	}

	return buffer;
}


/*******************************************************************************************/

FM::DSP::samplePtr& FM::DSP::abs::operator()(samplePtr& in) {
	if (in.getLength() != buffer.getLength()) buffer.resize(in.getLength());
	Flt * inPtr = in.getBuffer();
	Flt * outPtr = buffer.getBuffer();
	UInt length = in.getLength();

	while (length--) {
		Flt f = *inPtr++;
		*outPtr++ = (f >= 0 ? f : -f);
	}

	return buffer;
}